import pandas as pd

from sklearn.model_selection import StratifiedKFold

from . import BaseTask

class BaseCrossValidationTask(BaseTask):
  """
  The :class:`.BaseCrossValidationTask` is an abstract base class for implementing tasks that use cross-validation functionality.

  It can support both stratified k-fold cross-validation as well as cross-validation with user-defined folds.

  By default, stratified k-fold cross-validation is used with the default parameters of :meth:`.stratified_folds` method.
  """
  def __init__ (self, type_name, args):
    """
    This is the constructor method and can be used from child :class:`.BaseCrossValidationTask` implementations.

    :param type_name: The task's type name. A common practice is to pass the name of the task's class.
    :type type_name: str

    :param args: The task's parameters. A common practice is to pass the parameters at the time of task's object creation. It is a dictionary of key-value pairs, where the key is the parameter name and the value is the parameter value.
    :type args: dict
    """

    super(BaseCrossValidationTask, self).__init__(type_name, args)

    self.stratified_folds()

  def custom_folds(self, folds_data, fold_column='fold'):
    """
    Optional method.

    Use cross-validation with user-defined custom folds.

    :param folds_data: The input folds data. It can be either a URL, a disk file path or a pandas DataFrame. The folds data contain the user-defined folds for the samples. If a URL or a disk file path is provided the data must be formatted with the same separation delimiter (comma for CSV, tab for TSV, etc.) as the one used in the input data set files provided to the task. The data must contain two columns and the first row must be the header. The first column is for the sample IDs and needs to be the same as the one used in the input data set files provided to the task. The second column is for the fold IDs (e.g., 1 through 5, A through D, etc.).
    :type folds_data: {str or pandas DataFrame}

    :param fold_column: The column name for the fold IDs. It defaults to 'fold'.
    :type fold_column: str, optional

    :return: The object instance itself.
    :rtype: :class:`.BaseCrossValidationTask`
    """

    options = self._filter_arguments(locals())

    self._update_arguments({ 'fold_options' : options})

    self._update_arguments({ 'fold_method': 'custom' })

    return self

  def stratified_folds(self, total_folds=3, shuffle=False):
    """
    Optional method.

    Use stratified k-fold cross-validation.

    The folds are made by preserving the percentage of samples for each class.

    :param total_folds: Number of folds. Must be at least 2. It defaults to 3.
    :type total_folds: int, optional

    :param shuffle: Whether to shuffle each class's samples before splitting into batches. Note that the samples within each split will not be shuffled. It defaults to False.
    :type shuffle: bool, optional

    :return: The object instance itself.
    :rtype: :class:`.BaseCrossValidationTask`
    """

    options = self._filter_arguments(locals())

    self._update_arguments({ 'fold_options' : options})

    self._update_arguments({ 'fold_method': 'stratified' })

    return self

  def _build_cv_splits (self, X, y):
    if self.fold_method == 'custom':
      folds_data = self.fold_options['folds_data']

      if isinstance(folds_data, str):
        folds_data_frame = pd.read_csv(folds_data, delimiter=self.field_delimiter)
      else:
        folds_data_frame = folds_data.copy()

        folds_data_frame.reset_index(inplace=True, drop=True)

      return self._get_cv_splits(self.train_data_set_data_frame.merge(folds_data_frame, how='inner', on=self.id_column))
    else:
      return StratifiedKFold(n_splits=self.fold_options['total_folds'], shuffle=self.fold_options['shuffle'], random_state=self.random_seed if self.fold_options['shuffle'] else None).split(X, y)

  def _get_cv_splits(self, data_set_data_frame_with_folds):
    cv = []

    for fold_id in sorted(data_set_data_frame_with_folds[self.fold_options['fold_column']].unique()):
      train_indexes = data_set_data_frame_with_folds[data_set_data_frame_with_folds[self.fold_options['fold_column']] != fold_id].index.values

      validation_indexes = data_set_data_frame_with_folds[data_set_data_frame_with_folds[self.fold_options['fold_column']] == fold_id].index.values

      cv.append((train_indexes, validation_indexes))

    return cv


import copy

from stringcase import snakecase

from abc import ABC, abstractmethod

class BaseTask(ABC):
  """
  The :class:`.BaseTask` is an abstract base class for implementing tasks.

  A task is a configurable and reproducible piece of code built on top of scikit-learn that can be used in machine learning pipelines.
  """
  def __init__ (self, type_name, args):
    """
    This is the constructor method and can be used from child :class:`.BaseTask` implementations.

    :param type_name: The task's type name. A common practice is to pass the name of the task's class.
    :type type_name: str

    :param args: The task's parameters. A common practice is to pass the parameters at the time of task's object creation. It is a dictionary of key-value pairs, where the key is the parameter name and the value is the parameter value.
    :type args: dict
    """

    self.arguments = {}

    self._update_arguments({'type' : snakecase(type_name) })

    self._update_arguments(args)

    super(BaseTask, self).__init__()

  def get_type(self):
    """
    Get the task's type name.

    :return: The task's type name.
    :rtype: str
    """
    return self.type

  def get_configuration(self):
    """
    Get the task's parameters.

    :return: The task's parameters as a dictionary of key-value pairs, where the key is the parameter name and the value is the parameter value.
    :rtype: dict
    """
    return copy.deepcopy(self.arguments)

  @abstractmethod
  def run(self, output_directory):
    """
    An abstract method for running the task.

    :param output_directory: The output directory path under which task-related generated files are stored.
    :type output_directory: str
    """
    pass

  def _update_arguments (self, args):
    arguments = self._filter_arguments(args)

    self.arguments.update(arguments)

    self.__dict__.update(arguments)

  def _filter_arguments(self, args):
    return copy.deepcopy({ k:v for k,v in args.items() if not k.startswith("__") and not k == "self" and not k.endswith("__") })


import os, copy

import pandas as pd

from . import BaseTask

class PredictionTask(BaseTask):
  """
  The :class:`.PredictionTask` class can be used to predict new data using a scikit-learn estimator/pipeline.
  """
  def __init__ (self, estimator, data_set, field_delimiter=',', feature_columns='all', id_column='id', prediction_column='prediction', threshold=0.5):
    """
    This is the constructor method and can be used to create a new object instance of :class:`.PredictionTask` class.

    :param estimator: It can be either an estimator (e.g., LogisticRegression) or a pipeline ending with an estimator. The estimator needs to be able to predict probabilities through a ``predict_proba`` method.
    :type estimator: scikit-learn {estimator, pipeline}

    :param data_set: The input data set. It can be either a URL, a disk file path or a pandas DataFrame.
    :type data_set: {str, pandas DataFrame}

    :param field_delimiter: The separation delimiter (comma for CSV, tab for TSV, etc.) used in the input data set file. It defaults to ','.
    :type field_delimiter: str, optional

    :param feature_columns: Either 'all' to use from the input data set file all the columns or a list of column names to select specific columns. It defaults to 'all'.
    :type feature_columns: {str, list}, optional

    :param id_column: The name of the column in the input data set file containing the sample IDs. It defaults to 'id'.
    :type id_column: str, optional

    :param prediction_column: The name of the column for the predicted binary class labels. It defaults to 'prediction'.
    :type prediction_column: str, optional

    :param threshold: The threshold to use for converting the predicted probability into a binary class label. It defaults to 0.5.
    :type threshold: float, optional
    """

    arguments = copy.deepcopy(locals())

    super(PredictionTask, self).__init__(PredictionTask.__name__, arguments)

  def run(self, output_directory):
    """
    Run the task.

    The predictions are returned as a result and also stored in a *predictions.csv* CSV file under the output directory path.

    :param output_directory: The output directory path under which task-related generated files are stored.
    :type output_directory: str

    :return: The task's result. Specifically, the predictions for the input data set, containing the sample IDs, the predicted binary class labels, and the predicted probabilities for the positive class.
    :rtype: pandas DataFrame
    """

    if isinstance(self.data_set, str):
      data_set_data_frame = pd.read_csv(self.data_set, delimiter=self.field_delimiter)
    else:
      data_set_data_frame = self.data_set.copy()

      data_set_data_frame.reset_index(inplace=True, drop=True)

    ids = data_set_data_frame[self.id_column]

    X = data_set_data_frame.drop(columns=[self.id_column])

    if self.feature_columns != 'all':
      X = X[self.feature_columns]

    predictions = pd.DataFrame({ self.id_column : ids, self.prediction_column : self._calculate_y_hat_for_threshold(self.estimator, X, self.threshold), "probability" : self._calculate_probability(self.estimator, X)})

    predictions.to_csv(os.path.join(output_directory, f'predictions.csv'), index=False)

    return predictions

  def _calculate_y_hat_for_threshold (self, estimator, X, threshold):
    y_proba = estimator.predict_proba(X)

    y_hat = y_proba[:, 1] >= threshold

    return y_hat.astype(int)

  def _calculate_probability(self, estimator, X):
    y_proba = estimator.predict_proba(X)

    return y_proba[:, 1]


import os, copy

import pandas as pd

import featuretools as ft

from . import BaseTask

class DeepFeatureSynthesisTask(BaseTask):
  """
  The :class:`.DeepFeatureSynthesisTask` class is a wrapper for Featuretools. It can be used to automate feature engineering and create features from temporal and relational datasets.
  """
  def __init__ (self, entities=None,
                      relationships=None,
                      entityset=None,
                      target_entity=None,
                      cutoff_time=None,
                      instance_ids=None,
                      agg_primitives=None,
                      trans_primitives=None,
                      groupby_trans_primitives=None,
                      allowed_paths=None,
                      max_depth=2,
                      ignore_entities=None,
                      ignore_variables=None,
                      primitive_options=None,
                      seed_features=None,
                      drop_contains=None,
                      drop_exact=None,
                      where_primitives=None,
                      max_features=-1,
                      cutoff_time_in_index=False,
                      save_progress=None,
                      training_window=None,
                      approximate=None,
                      chunk_size=None,
                      n_jobs=1,
                      dask_kwargs=None,
                      verbose=False,
                      return_variable_types=None,
                      progress_callback=None,
                      include_cutoff_time=True,
                      export_feature_graphs=False,
                      export_feature_information=False,
                      id_column='id',
                      label_column='label'):
    """
    This is the constructor method and can be used to create a new object instance of :class:`.DeepFeatureSynthesisTask` class.

    Most of the arguments are documented here: https://featuretools.alteryx.com/en/stable/generated/featuretools.dfs.html#featuretools.dfs

    :param export_feature_graphs: If this task will export feature computation graphs. It defaults to False.
    :type export_feature_graphs: bool, optional

    :param export_feature_information: If this task will export feature information. The feature definitions can be used to recalculate features for a different data set. It defaults to False.
    :type export_feature_information: bool, optional

    :param id_column: The name of the column containing the sample IDs. It defaults to 'id'.
    :type id_column: str, optional

    :param label_column: The name of the column containing the ground truth labels. It defaults to 'label'.
    :type label_column: str, optional
    """
    arguments = copy.deepcopy(locals())

    super(DeepFeatureSynthesisTask, self).__init__(DeepFeatureSynthesisTask.__name__, arguments)

  def run(self, output_directory):
    """
    Run the task.

    The synthesized output dataset is returned as a result and also stored in a *synthesized_dataset.csv* file under the output directory path.

    The features information are stored in a *feature_information.html* file as a static HTML table under the output directory path.

    The feature computation graphs are stored as PNG files under the output directory path.

    Also, the feature definitions are stored in a *feature_definitions.txt* file under the output directory path.

    :param output_directory: The output directory path under which task-related generated files are stored.
    :type output_directory: str

    :return: The task's result. Specifically, **1)** ``synthesized_dataset``: The synthesized output dataset as a pandas DataFrame. **2)** ``feature_definitions``: The definitions of features in the synthesized output dataset. The feature definitions can be used to recalculate features for a different data set.
    :rtype: dict
    """
    synthesized_dataset, feature_defs = ft.dfs(
      entities=self.entities,
      relationships=self.relationships,
      entityset=self.entityset,
      target_entity=self.target_entity,
      cutoff_time=self.cutoff_time,
      instance_ids=self.instance_ids,
      agg_primitives=self.agg_primitives,
      trans_primitives=self.trans_primitives,
      groupby_trans_primitives=self.groupby_trans_primitives,
      allowed_paths=self.allowed_paths,
      max_depth=self.max_depth,
      ignore_entities=self.ignore_entities,
      ignore_variables=self.ignore_variables,
      primitive_options=self.primitive_options,
      seed_features=self.seed_features,
      drop_contains=self.drop_contains,
      drop_exact=self.drop_exact,
      where_primitives=self.where_primitives,
      max_features=self.max_features,
      cutoff_time_in_index=self.cutoff_time_in_index,
      save_progress=self.save_progress,
      training_window=self.training_window,
      approximate=self.approximate,
      chunk_size=self.chunk_size,
      n_jobs=self.n_jobs,
      dask_kwargs=self.dask_kwargs,
      verbose=self.verbose,
      return_variable_types=self.return_variable_types,
      progress_callback=self.progress_callback,
      include_cutoff_time=self.include_cutoff_time
    )

    label_related_columns_to_drop = []

    for column in synthesized_dataset:
      if column == self.label_column:
        pass
      elif self.label_column in column:
          label_related_columns_to_drop.append(column)

    synthesized_dataset = synthesized_dataset[[o for o in synthesized_dataset if o not in label_related_columns_to_drop]]

    feature_defs = [ o for o in feature_defs if o.get_name() != self.label_column and o.get_name() not in label_related_columns_to_drop ]

    features_df = pd.DataFrame({ 'feature_id' : range(len(feature_defs)), 'feature_def' : feature_defs })

    features_df['feature_name'] = features_df['feature_def'].apply(lambda o: o.get_name())
    features_df['feature_type'] = features_df['feature_def'].apply(lambda o: synthesized_dataset.dtypes[o.get_name()])

    if self.export_feature_graphs:
      features_df.apply(lambda o: ft.graph_feature(o['feature_def'], to_file=os.path.join(output_directory, 'feature_graphs', f'{o["feature_id"]}.png'), description=True), axis=1)

    if self.export_feature_information:
      features_df['feature_description'] = features_df['feature_def'].apply(lambda o: ft.describe_feature(o))

      features_df.drop(columns=['feature_def'], inplace=True)

      features_df.to_html(os.path.join(output_directory, 'feature_information.html'), index=False)

    synthesized_dataset.reset_index(inplace=True)

    synthesized_dataset.to_csv(os.path.join(output_directory, 'synthesized_dataset.csv'), index=False)

    ft.save_features(feature_defs, os.path.join(output_directory, 'feature_definitions.txt'))

    return { 'synthesized_dataset' : synthesized_dataset, 'feature_definitions': feature_defs }


import os, copy

import pandas as pd

import numpy as np

from math import floor

from functools import reduce

from plotly.offline import plot

from plotly import express

from . import BaseCrossValidationTask

class EvaluationCrossValidationTask(BaseCrossValidationTask):
  """
  The :class:`.EvaluationCrossValidationTask` class can be used to evaluate a scikit-learn estimator/pipeline on some data.

  The following evaluation results can be generated on-demand for hold-out test data set as well as train/validation cross-validation folds:

  * PR / ROC Curves
  * Confusion Matrixes
  * Classification Reports
  * Performance Metrics
  * False Positives
  * False Negatives

  It can support both stratified k-fold cross-validation as well as cross-validation with user-defined folds.

  By default, stratified k-fold cross-validation is used with the default parameters of :meth:`.stratified_folds` method.
  """
  def __init__ (self, estimator, train_data_set, test_data_set=None, estimator_params=None, field_delimiter=',', feature_columns='all', id_column='id', label_column='label', random_seed=42, threshold_selection_by='f1', metric_greater_is_better=True, threshold_tuning_range=(0.01, 1.0, 0.01), export_classification_reports=False, export_confusion_matrixes=False, export_roc_curves=False, export_pr_curves=False, export_false_positives_reports=False, export_false_negatives_reports=False, export_also_for_train_folds=False, fscore_beta=1):
    """
    This is the constructor method and can be used to create a new object instance of :class:`.EvaluationCrossValidationTask` class.

    :param estimator: It can be either an estimator (e.g., LogisticRegression) or a pipeline ending with an estimator. The estimator needs to be able to predict probabilities through a ``predict_proba`` method.
    :type estimator: scikit-learn {estimator, pipeline}

    :param train_data_set: The input train data set. It can be either a URL, a disk file path or a pandas DataFrame.
    :type train_data_set: {str or pandas DataFrame}

    :param test_data_set: The input test data set. It can be either a URL, a disk file path or a pandas DataFrame. It defaults to None.
    :type test_data_set: {str or pandas DataFrame}, optional

    :param estimator_params: The parameters to override in the provided estimator/pipeline. It defaults to None.
    :type estimator_params: dict, optional

    :param field_delimiter: The separation delimiter (comma for CSV, tab for TSV, etc.) used in the input train/test data set files. It defaults to ','.
    :type field_delimiter: str, optional

    :param feature_columns: Either 'all' to use from the input train/test data set files all the columns or a list of column names to select specific columns. It defaults to 'all'.
    :type feature_columns: {str, list}, optional

    :param id_column: The name of the column in the input train/test data set files containing the sample IDs. It defaults to 'id'.
    :type id_column: str, optional

    :param label_column: The name of the column in the input train/test data set files containing the ground truth labels. It defaults to 'label'.
    :type label_column: str, optional

    :param random_seed: The random seed used in the random number generator. It can be used to reproduce the output. It defaults to 42.
    :type random_seed: int, optional

    :param threshold_selection_by: The evaluation results will be generated either for a specific provided threshold value (e.g., 0.49) or for the best threshold found from threshold tuning, based on a specific provided metric (e.g., 'f1', 'f0.55'). It defaults to 'f1'.
    :type threshold_selection_by: {str, float}, optional

    :param metric_greater_is_better: This flag will control the direction of searching of the best threshold and it depends on the provided metric in ``threshold_selection_by``. True, means that greater metric values is better and False means the opposite. It defaults to True.
    :type metric_greater_is_better: bool, optional

    :param threshold_tuning_range: A range in form (start_value, stop_value, step_size) for generating a sequence of threshold values in threshold tuning. It generates the sequence by incrementing the start value using the step size until it reaches the stop value. It defaults to (0.01, 1.0, 0.01).
    :type threshold_tuning_range: tuple, optional

    :param export_classification_reports: If this task will export classification reports. It defaults to False.
    :type export_classification_reports: bool, optional

    :param export_confusion_matrixes:  If this task will export confusion matrixes. It defaults to False.
    :type export_confusion_matrixes: bool, optional

    :param export_roc_curves:  If this task will export ROC curves. It defaults to False.
    :type export_roc_curves: bool, optional

    :param export_pr_curves: If this task will export PR curves. It defaults to False.
    :type export_pr_curves: bool, optional

    :param export_false_positives_reports: If this task will export false positives reports. It defaults to False.
    :type export_false_positives_reports: bool, optional

    :param export_false_negatives_reports: If this task will export false negatives reports. It defaults to False.
    :type export_false_negatives_reports: bool, optional

    :param export_also_for_train_folds: If this task will export the evaluation results also for the train folds of cross-validation. It defaults to False.
    :type export_also_for_train_folds: bool, optional

    :param fscore_beta: The beta parameter in F-measure. It determines the weight of recall in the score. *beta < 1* lends more weight to precision, while *beta > 1* favors recall (*beta -> 0* considers only precision, *beta -> +inf* only recall). It defaults to 1.
    :type fscore_beta: float, optional
    """

    super(EvaluationCrossValidationTask, self).__init__(EvaluationCrossValidationTask.__name__, locals())

    pd.set_option('display.max_colwidth', None)

    self.fscore_beta_text = f'f{str(self.fscore_beta)}'

    self.train_text = 'train'

    self.validation_text = 'validation'

    self.test_text = 'test'

  def run(self, output_directory):
    """
    Run the task.

    All of the evaluation results are stored as files under the output directory path.

    :param output_directory: The output directory path under which task-related generated files are stored.
    :type output_directory: str

    :return: The task's result. Specifically, the threshold used along with its related performance metrics and summary metrics from all cross-validation splits as well as hold-out test data set.
    :rtype: dict
    """

    if isinstance(self.train_data_set, str):
      self.train_data_set_data_frame = pd.read_csv(self.train_data_set, delimiter=self.field_delimiter)
    else:
      self.train_data_set_data_frame = self.train_data_set.copy()

      self.train_data_set_data_frame.reset_index(inplace=True, drop=True)

    train_ids = self.train_data_set_data_frame[self.id_column]
    train_y = self.train_data_set_data_frame[self.label_column]
    train_X = self.train_data_set_data_frame.drop(columns=[self.label_column, self.id_column])
    if self.feature_columns != 'all': train_X = train_X[self.feature_columns]

    labels = sorted(train_y.unique())
    if len(labels) != 2: raise Exception('The unique values for the y target should be exactly two.')

    splits_threshold_metrics = self._get_splits_threshold_metrics(train_X, train_y, self._build_cv_splits(train_X, train_y))
    splits_threshold_metrics.to_html(os.path.join(output_directory, 'cv_splits_threshold_metrics.html'), index=False)
    splits_threshold_metrics_summary = self._get_splits_threshold_metrics_summary(splits_threshold_metrics)
    splits_threshold_metrics_summary.to_html(os.path.join(output_directory, 'cv_splits_threshold_metrics_summary.html'), index=False)

    self.threshold, threshold_metrics = self._get_threshold_and_its_metrics(splits_threshold_metrics_summary)
    splits_y_and_y_hat_validation, splits_y_and_y_hat_train = self._get_splits_y_and_y_hat_results(train_X, train_y, train_ids, self._build_cv_splits(train_X, train_y))

    result = { 'threshold' : self.threshold, 'cv_threshold_metrics' : threshold_metrics, 'cv_splits_threshold_metrics' : splits_threshold_metrics, 'cv_splits_threshold_metrics_summary' : splits_threshold_metrics_summary }

    y_and_y_hat_test = None
    test_threshold_metrics = None

    if self.test_data_set is not None:
      if isinstance(self.test_data_set, str):
        self.test_data_set_data_frame = pd.read_csv(self.test_data_set, delimiter=self.field_delimiter)
      else:
        self.test_data_set_data_frame = self.test_data_set.copy()

        self.test_data_set_data_frame.reset_index(inplace=True, drop=True)

      test_ids = self.test_data_set_data_frame[self.id_column]
      test_y = self.test_data_set_data_frame[self.label_column]
      test_X = self.test_data_set_data_frame.drop(columns=[self.label_column, self.id_column])
      if self.feature_columns != 'all': test_X = test_X[self.feature_columns]

      np.random.seed(self.random_seed)
      estimator = self._build_estimator()
      estimator.fit(train_X, train_y)

      test_threshold_metrics = self._get_threshold_metrics(estimator, test_X, test_y, self.test_text)
      test_threshold_metrics.to_html(os.path.join(output_directory, 'test_threshold_metrics.html'), index=False)
      y_and_y_hat_test = self._get_y_and_y_hat_results(estimator, test_X, test_y, test_ids)

      result.update({'test_threshold_metrics' : test_threshold_metrics})

    if self.export_pr_curves: self._build_all_precision_recall_curves(splits_threshold_metrics, splits_threshold_metrics_summary, self._create_directory_path(output_directory, 'pr_curves'), test_threshold_metrics)
    if self.export_roc_curves: self._build_all_roc_curves(splits_threshold_metrics, splits_threshold_metrics_summary, self._create_directory_path(output_directory, 'roc_curves'), test_threshold_metrics)
    if self.export_confusion_matrixes: self._build_all_confusion_matrixes(splits_y_and_y_hat_validation, splits_y_and_y_hat_train, labels, self._create_directory_path(output_directory, 'confusion_matrixes'), y_and_y_hat_test)
    if self.export_classification_reports: self._build_all_classification_reports(splits_y_and_y_hat_validation, splits_y_and_y_hat_train, labels, self._create_directory_path(output_directory, 'classification_reports'), y_and_y_hat_test)
    if self.export_false_positives_reports: self._build_all_false_positives_reports(splits_y_and_y_hat_validation, splits_y_and_y_hat_train, self._create_directory_path(output_directory, 'false_positives_reports'), y_and_y_hat_test)
    if self.export_false_negatives_reports: self._build_all_false_negatives_reports(splits_y_and_y_hat_validation, splits_y_and_y_hat_train, self._create_directory_path(output_directory, 'false_negatives_reports'), y_and_y_hat_test)

    return result

  def _get_threshold_and_its_metrics (self, splits_threshold_metrics_summary):
    if isinstance(self.threshold_selection_by, float):
      row = splits_threshold_metrics_summary.loc[splits_threshold_metrics_summary['threshold'] == self.threshold_selection_by].squeeze()
    elif isinstance(self.threshold_selection_by, str):
      mean_column = f'validation_{self.threshold_selection_by}_mean'

      std_column = f'validation_{self.threshold_selection_by}_std'

      if mean_column not in splits_threshold_metrics_summary.columns or std_column not in splits_threshold_metrics_summary.columns:
        raise Exception(f"The specified metric [{self.threshold_selection_by}] for threshold selection is invalid.")

      row = splits_threshold_metrics_summary.sort_values([mean_column, std_column], ascending = (not self.metric_greater_is_better, True)).head(1).squeeze()
    else:
      raise Exception(f"The specified threshold selection criteria [{self.threshold_selection_by}] is invalid.")

    if row.empty: raise Exception(f"For the specified threshold selection criteria [{self.threshold_selection_by}] no threshold can be selected.")

    return self._truncate_number(row['threshold'], 3), row.drop('threshold')

  def _build_all_false_negatives_reports(self, splits_y_and_y_hat_validation, splits_y_and_y_hat_train, output_directory, y_and_y_hat_test=None):
    self._build_false_negatives_reports(splits_y_and_y_hat_validation, output_directory, self.validation_text)

    if self.export_also_for_train_folds: self._build_false_negatives_reports(splits_y_and_y_hat_train, output_directory, self.train_text)

    if y_and_y_hat_test: self._build_false_negatives_report(y_and_y_hat_test[0], y_and_y_hat_test[1], y_and_y_hat_test[2], output_directory, f'for {self.test_text} data', f'{self.test_text}_data')

  def _build_false_negatives_reports(self, splits_y_and_y_hat, output_directory, data_split):
    for split_index, (y, y_hat, ids) in enumerate(splits_y_and_y_hat):
      self._build_false_negatives_report(y, y_hat, ids, output_directory, f'for CV split {split_index} {data_split} data', f'cv_split_{split_index}_{data_split}_data')

  def _build_false_negatives_report(self, y, y_hat, ids, output_directory, title_part_text, file_part_text):
    data = np.vstack((y, y_hat, ids)).T

    false_negatives = data[(data[:,0] == 1) & (data[:,1] == 0)][:,2]

    if false_negatives.size != 0:
      with open(os.path.join(output_directory, f'false_negatives_report_{file_part_text}.txt'), "w") as f: f.write(f"False negatives report {title_part_text} (Threshold = {self.threshold})" + '\n'*2 + '\n'.join(map(str, false_negatives)))

  def _build_all_false_positives_reports(self, splits_y_and_y_hat_validation, splits_y_and_y_hat_train, output_directory, y_and_y_hat_test=None):
    self._build_false_positives_reports(splits_y_and_y_hat_validation, output_directory, self.validation_text)

    if self.export_also_for_train_folds: self._build_false_positives_reports(splits_y_and_y_hat_train, output_directory, self.train_text)

    if y_and_y_hat_test: self._build_false_positives_report(y_and_y_hat_test[0], y_and_y_hat_test[1], y_and_y_hat_test[2], output_directory, f'for {self.test_text} data', f'{self.test_text}_data')

  def _build_false_positives_reports(self, splits_y_and_y_hat, output_directory, data_split):
    for split_index, (y, y_hat, ids) in enumerate(splits_y_and_y_hat):
      self._build_false_positives_report(y, y_hat, ids, output_directory, f'for CV split {split_index} {data_split} data', f'cv_split_{split_index}_{data_split}_data')

  def _build_false_positives_report(self, y, y_hat, ids, output_directory, title_part_text, file_part_text):
    data = np.vstack((y, y_hat, ids)).T

    false_positives = data[(data[:,0] == 0) & (data[:,1] == 1)][:,2]

    if false_positives.size != 0:
      with open(os.path.join(output_directory, f'false_positives_report_{file_part_text}.txt'), "w") as f: f.write(f"False positives report {title_part_text} (Threshold = {self.threshold})" + '\n'*2 + '\n'.join(map(str, false_positives)))

  def _build_all_confusion_matrixes(self, splits_y_and_y_hat_validation, splits_y_and_y_hat_train, labels, output_directory, y_and_y_hat_test=None):
    self._build_confusion_matrixes(splits_y_and_y_hat_validation, output_directory, self.validation_text, labels)

    if self.export_also_for_train_folds: self._build_confusion_matrixes(splits_y_and_y_hat_train, output_directory, self.train_text, labels)

    if y_and_y_hat_test: self._build_confusion_matrix(y_and_y_hat_test[0], y_and_y_hat_test[1], output_directory, f'for {self.test_text} data', f'{self.test_text}_data', labels)

  def _build_confusion_matrixes(self, splits_y_and_y_hat, output_directory, data_split, labels):
    for split_index, (y, y_hat, _) in enumerate(splits_y_and_y_hat):
      self._build_confusion_matrix(y, y_hat, output_directory, f'for CV split {split_index} {data_split} data', f'cv_split_{split_index}_{data_split}_data', labels)

    y = reduce(lambda x, y: np.concatenate((x, y)), map(lambda x: x[0], splits_y_and_y_hat))
    y_hat = reduce(lambda x, y: np.concatenate((x, y)), map(lambda x: x[1], splits_y_and_y_hat))

    self._build_confusion_matrix(y, y_hat, output_directory, f'for CV {data_split} data', f'cv_{data_split}_data', labels)

  def _build_confusion_matrix(self, y, y_hat, output_directory, title_part_text, file_part_text, labels):
    import matplotlib.pyplot as plt

    from scikitplot.metrics import plot_confusion_matrix

    plot_confusion_matrix(y, y_hat, labels=labels, title=f'Confusion matrix {title_part_text} (Threshold = {self.threshold})')

    plt.savefig(os.path.join(output_directory, f'confusion_matrix_{file_part_text}.png'))

    plt.close()

  def _build_all_classification_reports(self, splits_y_and_y_hat_validation, splits_y_and_y_hat_train, labels, output_directory, y_and_y_hat_test=None):
    truncate_precision = 3

    self._build_classification_reports(splits_y_and_y_hat_validation, output_directory, self.validation_text, labels, truncate_precision)

    if self.export_also_for_train_folds: self._build_classification_reports(splits_y_and_y_hat_train, output_directory, self.train_text, labels, truncate_precision)

    if y_and_y_hat_test: self._build_classification_report(y_and_y_hat_test[0], y_and_y_hat_test[1], output_directory, f'for {self.test_text} data', f'{self.test_text}_data', labels, truncate_precision)

  def _build_classification_reports(self, splits_y_and_y_hat, output_directory, data_split, labels, truncate_precision):
    for split_index, (y, y_hat, _) in enumerate(splits_y_and_y_hat):
      self._build_classification_report(y, y_hat, output_directory, f'for CV split {split_index} {data_split} data', f'cv_split_{split_index}_{data_split}_data', labels, truncate_precision)

    y = reduce(lambda x, y: np.concatenate((x, y)), map(lambda x: x[0], splits_y_and_y_hat))
    y_hat = reduce(lambda x, y: np.concatenate((x, y)), map(lambda x: x[1], splits_y_and_y_hat))

    self._build_classification_report(y, y_hat, output_directory, f'for CV {data_split} data', f'cv_{data_split}_data', labels, truncate_precision)

  def _build_classification_report(self, y, y_hat, output_directory, title_part_text, file_part_text, labels, truncate_precision):
    from functools import partial

    from sklearn.metrics import precision_recall_fscore_support

    summary_metrics = list(precision_recall_fscore_support(y_true=y, y_pred=y_hat, labels=labels, beta=self.fscore_beta))

    average_metrics = list(precision_recall_fscore_support(y_true=y, y_pred=y_hat, labels=labels, beta=self.fscore_beta, average='weighted'))

    classification_report = pd.DataFrame(summary_metrics, index=['precision', 'recall', self.fscore_beta_text, 'support'])

    support = classification_report.loc['support']

    total = support.sum()

    average_metrics[-1] = total

    classification_report['avg / total'] = average_metrics

    classification_report = classification_report.T

    classification_report = classification_report.applymap(partial(self._truncate_number, truncate_precision=truncate_precision))

    classification_report['support'] = classification_report['support'].map(int)

    with open(os.path.join(output_directory, f'classification_report_{file_part_text}.txt'), "w") as f: f.write(f"Classification report {title_part_text} (Threshold = {self.threshold})" + '\n'*2 + f"{classification_report}")

  def _truncate_number(self, number, truncate_precision):
    return floor(number * 10 ** truncate_precision) / 10 ** truncate_precision

  def _build_all_precision_recall_curves(self, splits_threshold_metrics, splits_threshold_metrics_summary, output_directory, test_threshold_metrics=None):
    self._build_cv_precision_recall_curves(splits_threshold_metrics[['threshold', 'cv_split', 'validation_precision', 'validation_recall', f'validation_{self.fscore_beta_text}']], output_directory, self.validation_text)
    if self.export_also_for_train_folds: self._build_cv_precision_recall_curves(splits_threshold_metrics[['threshold', 'cv_split', 'train_precision', 'train_recall', f'train_{self.fscore_beta_text}']], output_directory, self.train_text)

    self._build_average_precision_recall_curve(splits_threshold_metrics_summary[['threshold', 'validation_precision_mean', 'validation_recall_mean', f'validation_{self.fscore_beta_text}_mean','validation_precision_std', 'validation_recall_std', f'validation_{self.fscore_beta_text}_std']], output_directory, self.validation_text)
    if self.export_also_for_train_folds: self._build_average_precision_recall_curve(splits_threshold_metrics_summary[['threshold', 'train_precision_mean', 'train_recall_mean', f'train_{self.fscore_beta_text}_mean', 'train_precision_std', 'train_recall_std', f'train_{self.fscore_beta_text}_std']], output_directory, self.train_text)

    if test_threshold_metrics is not None: self._build_precision_recall_curve(test_threshold_metrics[['threshold', 'test_precision', 'test_recall', f'test_{self.fscore_beta_text}']], output_directory, self.test_text)

  def _build_all_roc_curves(self, splits_threshold_metrics, splits_threshold_metrics_summary, output_directory, test_threshold_metrics=None):
    self._build_cv_roc_curves(splits_threshold_metrics[['threshold', 'cv_split', 'validation_true_positive_rate', 'validation_false_positive_rate']], output_directory, self.validation_text)
    if self.export_also_for_train_folds: self._build_cv_roc_curves(splits_threshold_metrics[['threshold', 'cv_split', 'train_true_positive_rate', 'train_false_positive_rate']], output_directory, self.train_text)

    self._build_average_roc_curve(splits_threshold_metrics_summary[['threshold', 'validation_true_positive_rate_mean', 'validation_false_positive_rate_mean', 'validation_true_positive_rate_std', 'validation_false_positive_rate_std']], output_directory, self.validation_text)
    if self.export_also_for_train_folds: self._build_average_roc_curve(splits_threshold_metrics_summary[['threshold', 'train_true_positive_rate_mean', 'train_false_positive_rate_mean', 'train_true_positive_rate_std', 'train_false_positive_rate_std']], output_directory, self.train_text)

    if test_threshold_metrics is not None: self._build_roc_curve(test_threshold_metrics[['threshold', 'test_true_positive_rate', 'test_false_positive_rate']], output_directory, self.test_text)

  def _build_average_precision_recall_curve(self, data_frame, output_directory, data_split):
    data_frame.columns = data_frame.columns.str.replace(f'^{data_split}_', '')

    fig = express.line(data_frame, x="recall_mean", y="precision_mean", hover_data=['threshold', f'{self.fscore_beta_text}_mean', 'precision_std', 'recall_std', f'{self.fscore_beta_text}_std'], title=f"Mean Precision-Recall curve from cross validation (CV) splits in {data_split} data")

    self._change_styles_in_figure(fig, x_title="Recall (mean)", y_title="Precision (mean)")

    plot(fig, filename=os.path.join(output_directory, f'cv_mean_pr_curve_{data_split}_data.html'), auto_open=False)

  def _build_average_roc_curve(self, data_frame, output_directory, data_split):
    data_frame.columns = data_frame.columns.str.replace(f'^{data_split}_', '')

    fig = express.line(data_frame, x="false_positive_rate_mean", y="true_positive_rate_mean", hover_data=['threshold', 'false_positive_rate_std', 'true_positive_rate_std'], title=f"Mean ROC curve from cross validation (CV) splits in {data_split} data")

    self._change_styles_in_figure(fig, x_title="False Positive Rate (mean)", y_title="True Positive Rate (mean)")

    plot(fig, filename=os.path.join(output_directory, f'cv_mean_roc_curve_{data_split}_data.html'), auto_open=False)

  def _build_cv_precision_recall_curves(self, data_frame, output_directory, data_split):
    data_frame.columns = data_frame.columns.str.replace(f'^{data_split}_', '')

    fig = express.line(data_frame, x="recall", y="precision", hover_data=['threshold', self.fscore_beta_text], color="cv_split", title=f"Precision-Recall curves from cross validation (CV) splits in {data_split} data")

    self._change_styles_in_figure(fig, x_title="Recall", y_title="Precision")

    plot(fig, filename=os.path.join(output_directory, f'cv_pr_curves_{data_split}_data.html'), auto_open=False)

  def _build_cv_roc_curves(self, data_frame, output_directory, data_split):
    data_frame.columns = data_frame.columns.str.replace(f'^{data_split}_', '')

    fig = express.line(data_frame, x="false_positive_rate", y="true_positive_rate", hover_data=['threshold'], color="cv_split", title=f"ROC curves from cross validation (CV) splits in {data_split} data")

    self._change_styles_in_figure(fig, x_title="False Positive Rate", y_title="True Positive Rate")

    plot(fig, filename=os.path.join(output_directory, f'cv_roc_curves_{data_split}_data.html'), auto_open=False)

  def _build_roc_curve(self, data_frame, output_directory, data_split):
    data_frame.columns = data_frame.columns.str.replace(f'^{data_split}_', '')

    fig = express.line(data_frame, x="false_positive_rate", y="true_positive_rate", hover_data=['threshold'], title=f"ROC curve for {data_split} data")

    self._change_styles_in_figure(fig, x_title="False Positive Rate", y_title="True Positive Rate")

    plot(fig, filename=os.path.join(output_directory, f'roc_curve_{data_split}_data.html'), auto_open=False)

  def _build_precision_recall_curve(self, data_frame, output_directory, data_split):
    data_frame.columns = data_frame.columns.str.replace(f'^{data_split}_', '')

    fig = express.line(data_frame, x="recall", y="precision", hover_data=['threshold', self.fscore_beta_text], title=f"Precision-Recall curve for {data_split} data")

    self._change_styles_in_figure(fig, x_title="Recall", y_title="Precision")

    plot(fig, filename=os.path.join(output_directory, f'pr_curve_{data_split}_data.html'), auto_open=False)

  def _change_styles_in_figure (self, figure, x_title, y_title):
    figure.update_traces(mode='lines+markers')

    figure.update_layout(
      plot_bgcolor = 'rgba(0,0,0,0)',
      xaxis_title = x_title,
      yaxis_title = y_title,
      xaxis_gridcolor = 'rgb(159, 197, 232)',
      yaxis_gridcolor = 'rgb(159, 197, 232)'
    )

  def _create_directory_path(self, output_directory, directory_name):
    directory_path = os.path.join(output_directory, directory_name)

    os.makedirs(directory_path, exist_ok=True)

    return directory_path

  def _calculate_y_hat_for_threshold (self, estimator, X, threshold):
    y_proba = estimator.predict_proba(X)

    y_hat = y_proba[:, 1] >= threshold

    return y_hat.astype(int)

  def _build_estimator (self):
    estimator = copy.deepcopy(self.estimator)

    if self.estimator_params: estimator.set_params(**self.estimator_params)

    return estimator

  def _calculate_diagnostic_performance (self, y, y_hat, data_split):
    true_positives = np.sum(np.logical_and(y_hat == 1, y == 1))
    true_negatives = np.sum(np.logical_and(y_hat == 0, y == 0))

    false_positives = np.sum(np.logical_and(y_hat == 1, y == 0))
    false_negatives = np.sum(np.logical_and(y_hat == 0, y == 1))

    accuracy = (true_positives + true_negatives) / (true_positives + true_negatives + false_positives + false_negatives)

    true_positive_rate = true_positives / (true_positives + false_negatives)
    false_positive_rate = false_positives / (false_positives + true_negatives)

    sensitivity = true_positive_rate
    specificity = 1 - false_positive_rate

    geometric_mean = np.sqrt(sensitivity * specificity)

    precision = true_positives / (true_positives + false_positives)
    recall = sensitivity

    fbeta_score = self._calculate_fbeta_score(precision, recall, self.fscore_beta)

    performance = {}

    performance[f'{data_split}_sensitivity'] = sensitivity
    performance[f'{data_split}_specificity'] = specificity

    performance[f'{data_split}_accuracy'] = accuracy

    performance[f'{data_split}_true_positive_rate'] = true_positive_rate
    performance[f'{data_split}_false_positive_rate'] = false_positive_rate
    performance[f'{data_split}_geometric_mean'] = geometric_mean

    performance[f'{data_split}_precision'] = precision
    performance[f'{data_split}_recall'] = recall
    performance[f'{data_split}_{self.fscore_beta_text}'] = fbeta_score

    return performance

  def _calculate_fbeta_score (self, precision, recall, beta):
    bb = beta**2.0

    return (1 + bb) * (precision * recall) / (bb * precision + recall)

  def _get_splits_threshold_metrics_summary(self, dataframe):
    rows_with_nulls = dataframe[dataframe.isnull().any(axis=1)]

    threshold_values_to_remove = rows_with_nulls['threshold'].unique()

    rows_without_nulls = dataframe.loc[~dataframe['threshold'].isin(threshold_values_to_remove)]

    if rows_without_nulls.empty: raise Exception("The summary of threshold metrics cannot be generated.")

    rows_without_nulls = rows_without_nulls.drop(columns='cv_split')

    summary = rows_without_nulls.groupby('threshold').mean().add_suffix('_mean').merge(rows_without_nulls.groupby('threshold').std().add_suffix('_std'), left_index=True, right_index=True)

    return summary.reset_index()

  def _get_threshold_metrics_data_frame_columns(self, data_split):
    return [
      'threshold',

      f'{data_split}_sensitivity',
      f'{data_split}_specificity',
      f'{data_split}_accuracy',
      f'{data_split}_true_positive_rate',
      f'{data_split}_false_positive_rate',
      f'{data_split}_geometric_mean',
      f'{data_split}_precision',
      f'{data_split}_recall',
      f'{data_split}_{self.fscore_beta_text}'
    ]

  def _get_splits_threshold_metrics_data_frame_columns(self):
    columns = [
      'threshold',
      'cv_split',

      f'{self.validation_text}_sensitivity',
      f'{self.validation_text}_specificity',
      f'{self.validation_text}_accuracy',
      f'{self.validation_text}_true_positive_rate',
      f'{self.validation_text}_false_positive_rate',
      f'{self.validation_text}_geometric_mean',
      f'{self.validation_text}_precision',
      f'{self.validation_text}_recall',
      f'{self.validation_text}_{self.fscore_beta_text}'
    ]

    if self.export_also_for_train_folds:
      columns.extend([
        f'{self.train_text}_sensitivity',
        f'{self.train_text}_specificity',
        f'{self.train_text}_accuracy',
        f'{self.train_text}_true_positive_rate',
        f'{self.train_text}_false_positive_rate',
        f'{self.train_text}_geometric_mean',
        f'{self.train_text}_precision',
        f'{self.train_text}_recall',
        f'{self.train_text}_{self.fscore_beta_text}'
      ])

    return columns

  def _get_split_threshold_metrics (self, estimator, split_id, X_validation, y_validation, X_train, y_train):
    threshold_rows = []

    for threshold in np.arange (self.threshold_tuning_range[0], self.threshold_tuning_range[1], self.threshold_tuning_range[2]):
      row = [ threshold, split_id ]

      row.extend(self._calculate_diagnostic_performance(y_validation, self._calculate_y_hat_for_threshold(estimator, X_validation, threshold), self.validation_text).values())
      if self.export_also_for_train_folds: row.extend(self._calculate_diagnostic_performance(y_train, self._calculate_y_hat_for_threshold(estimator, X_train, threshold), self.train_text).values())

      threshold_rows.append(row)

    return pd.DataFrame(threshold_rows, columns=self._get_splits_threshold_metrics_data_frame_columns())

  def _get_splits_threshold_metrics (self, X, y, cv):
    np.random.seed(self.random_seed)

    splits_threshold_metrics = []

    for split_id, (train_indexes, validation_indexes) in enumerate(cv):
      X_train, y_train = X.loc[train_indexes], y.loc[train_indexes]
      X_validation, y_validation = X.loc[validation_indexes], y.loc[validation_indexes]

      estimator = self._build_estimator()
      estimator.fit(X_train, y_train)

      splits_threshold_metrics.append(self._get_split_threshold_metrics(estimator, split_id, X_validation, y_validation, X_train, y_train))

    splits_threshold_metrics = pd.concat(splits_threshold_metrics).reset_index(drop=True)

    return splits_threshold_metrics

  def _get_splits_y_and_y_hat_results (self, X, y, ids, cv):
    np.random.seed(self.random_seed)

    splits_y_and_y_hat_validation = []
    splits_y_and_y_hat_train = []

    for split_id, (train_indexes, validation_indexes) in enumerate(cv):
      X_train, y_train = X.loc[train_indexes], y.loc[train_indexes]
      X_validation, y_validation = X.loc[validation_indexes], y.loc[validation_indexes]

      ids_validation = ids.loc[validation_indexes]
      ids_train = ids.loc[train_indexes]

      estimator = self._build_estimator()
      estimator.fit(X_train, y_train)

      splits_y_and_y_hat_validation.append(self._get_y_and_y_hat_results(estimator, X_validation, y_validation, ids_validation))
      if self.export_also_for_train_folds: splits_y_and_y_hat_train.append(self._get_y_and_y_hat_results(estimator, X_train, y_train, ids_train))

    return splits_y_and_y_hat_validation, splits_y_and_y_hat_train

  def _get_y_and_y_hat_results (self, estimator, X, y, ids):
    return (y, self._calculate_y_hat_for_threshold(estimator, X, self.threshold), ids)

  def _get_threshold_metrics (self, estimator, X, y, data_split):
    threshold_rows = []

    for threshold in np.arange (self.threshold_tuning_range[0], self.threshold_tuning_range[1], self.threshold_tuning_range[2]):
      row = [ threshold ]

      row.extend(self._calculate_diagnostic_performance(y, self._calculate_y_hat_for_threshold(estimator, X, threshold), data_split).values())

      threshold_rows.append(row)

    return pd.DataFrame(threshold_rows, columns=self._get_threshold_metrics_data_frame_columns(data_split))


from abc import ABC, abstractmethod

class BaseNotifier(ABC):
    """
    The :class:`.BaseNotifier` is an abstract base class for implementing notifiers.

    A notifier can be used to send notifications.
    """
    @abstractmethod
    def notify(self, message):
        """
        An abstract method for sending the notification.

        :param message: The notification's message.
        :type message: str
        """
        pass


import os, copy

import pandas as pd

import numpy as np

from sklearn.feature_selection import RFECV

from sklearn.pipeline import Pipeline

from . import BaseCrossValidationTask

class FeatureSelectionCrossValidationTask(BaseCrossValidationTask):
  """
  The :class:`.FeatureSelectionCrossValidationTask` class can be used to perform feature selection with Recursive Feature Elimination using a scikit-learn estimator on some data.

  A scikit-learn preprocessor can be used on the input train data set before feature selection runs.

  It can support both stratified k-fold cross-validation as well as cross-validation with user-defined folds.

  By default, stratified k-fold cross-validation is used with the default parameters of :meth:`.stratified_folds` method.
  """
  def __init__ (self, estimator, train_data_set, estimator_params=None, field_delimiter=',', preprocessor=None, preprocessor_params=None, min_features_to_select=1, scoring='f1', feature_columns='all', id_column='id', label_column='label', random_seed=42, verbose=3, n_jobs=1):
    """
    This is the constructor method and can be used to create a new object instance of :class:`.FeatureSelectionCrossValidationTask` class.

    :param estimator: An estimator (e.g., LogisticRegression). It needs to provide feature importances through either a ``coef_`` or a ``feature_importances_`` attribute.
    :type estimator: scikit-learn estimator

    :param train_data_set: The input train data set. It can be either a URL, a disk file path or a pandas DataFrame.
    :type train_data_set: {str, pandas DataFrame}

    :param estimator_params: The parameters to override in the provided estimator. It defaults to None.
    :type estimator_params: dict, optional

    :param field_delimiter: The separation delimiter (comma for CSV, tab for TSV, etc.) used in the input train data set file. It defaults to ','.
    :type field_delimiter: str, optional

    :param preprocessor: The preprocessor you want to run on the input train data set before feature selection. You can set for example a scikit-learn ColumnTransformer, OneHotEncoder, etc. It defaults to None.
    :type preprocessor: scikit-learn preprocessor, optional

    :param preprocessor_params: The parameters to override in the provided preprocessor. It defaults to None.
    :type preprocessor_params: dict, optional

    :param min_features_to_select: The minimum number of features to be selected. This number of features will always be scored. It defaults to 1.
    :type min_features_to_select: int, optional

    :param scoring: A single scikit-learn scorer string (e.g., 'f1') or a callable that is built with scikit-learn ``make_scorer``. Note that when using custom scorers, each scorer should return a single value. It defaults to 'f1'.
    :type scoring: {str, callable}, optional

    :param feature_columns: Either 'all' to use from the input train data set file all the columns or a list of column names to select specific columns. It defaults to 'all'.
    :type feature_columns: {str, list}, optional

    :param id_column: The name of the column in the input train data set file containing the sample IDs. It defaults to 'id'.
    :type id_column: str, optional

    :param label_column: The name of the column in the input train data set file containing the ground truth labels. It defaults to 'label'.
    :type label_column: str, optional

    :param random_seed: The random seed used in the random number generator. It can be used to reproduce the output. It defaults to 42.
    :type random_seed: int, optional

    :param verbose: Controls the verbosity of output. The higher, the more messages. It defaults to 3.
    :type verbose: int, optional

    :param n_jobs: Number of jobs to run in parallel. -1 means using all processors. It defaults to 1.
    :type n_jobs: int, optional
    """

    super(FeatureSelectionCrossValidationTask, self).__init__(FeatureSelectionCrossValidationTask.__name__, locals())

  def run(self, output_directory):
    """
    Run the task.

    The selected features are returned as a result and also stored in a *features_selected.txt* file under the output directory path.

    :param output_directory: The output directory path under which task-related generated files are stored.
    :type output_directory: str
    
    :return: The task's result. Specifically, the selected features, which can be either column names from the input train data set or column indexes from the preprocessed data set, depending on whether a ``preprocessor`` was used or not.
    :rtype: list
    """
    if isinstance(self.train_data_set, str):
      self.train_data_set_data_frame = pd.read_csv(self.train_data_set, delimiter=self.field_delimiter)
    else:
      self.train_data_set_data_frame = self.train_data_set.copy()

      self.train_data_set_data_frame.reset_index(inplace=True, drop=True)

    y = self.train_data_set_data_frame[self.label_column]

    X = self.train_data_set_data_frame.drop(columns=[self.label_column, self.id_column])

    if self.feature_columns != 'all':
      X = X[self.feature_columns]

    np.random.seed(self.random_seed)

    model = RFECV(self._build_estimator(), min_features_to_select=self.min_features_to_select, step=1, cv=self._build_cv_splits(X, y), scoring=self.scoring, verbose=self.verbose, n_jobs=self.n_jobs)

    if self.preprocessor:
      model = Pipeline(steps=[('preprocessor', self._build_preprocessor()), ('selection', model)])

    model.fit(X, y)

    if self.preprocessor:
      features_selected = np.nonzero(model.named_steps.selection.support_)[0].tolist()
    else:
      features_selected = X.columns[model.support_].values.tolist()

    with open(os.path.join(output_directory, 'features_selected.txt'), "w") as f: f.writelines('\n'.join(map(str, features_selected)))

    return features_selected

  def _build_preprocessor (self):
    preprocessor = copy.deepcopy(self.preprocessor)

    if self.preprocessor_params: preprocessor.set_params(**self.preprocessor_params)

    return preprocessor

  def _build_estimator (self):
    estimator = copy.deepcopy(self.estimator)

    if self.estimator_params: estimator.set_params(**self.estimator_params)

    return estimator


import os, joblib, copy

import pandas as pd

import numpy as np

from . import BaseTask

class TrainTask(BaseTask):
  """
  The :class:`.TrainTask` class can be used to fit a scikit-learn estimator/pipeline on train data.
  """
  def __init__ (self, estimator, train_data_set, estimator_params=None, field_delimiter=',', feature_columns='all', id_column='id', label_column='label', random_seed=42):
    """
    This is the constructor method and can be used to create a new object instance of :class:`.TrainTask` class.

    :param estimator: It can be either an estimator (e.g., LogisticRegression) or a pipeline ending with an estimator.
    :type estimator: scikit-learn {estimator, pipeline}

    :param train_data_set: The input train data set. It can be either a URL, a disk file path or a pandas DataFrame.
    :type train_data_set: {str, pandas DataFrame}

    :param estimator_params: The parameters to override in the provided estimator/pipeline. It defaults to None.
    :type estimator_params: dict, optional

    :param field_delimiter: The separation delimiter (comma for CSV, tab for TSV, etc.) used in the input train data set file. It defaults to ','.
    :type field_delimiter: str, optional

    :param feature_columns: Either 'all' to use from the input train data set file all the columns or a list of column names to select specific columns. It defaults to 'all'.
    :type feature_columns: {str, list}, optional

    :param id_column: The name of the column in the input train data set file containing the sample IDs. It defaults to 'id'.
    :type id_column: str, optional

    :param label_column: The name of the column in the input train data set file containing the ground truth labels. It defaults to 'label'.
    :type label_column: str, optional

    :param random_seed: The random seed used in the random number generator. It can be used to reproduce the output. It defaults to 42.
    :type random_seed: int, optional
    """
    arguments = copy.deepcopy(locals())

    super(TrainTask, self).__init__(TrainTask.__name__, arguments)

  def run(self, output_directory):
    """
    Run the task.

    The fitted estimator/pipeline is returned as a result and also stored in a *trained_model.pkl* pickle file under the output directory path.

    :param output_directory: The output directory path under which task-related generated files are stored.
    :type output_directory: str

    :return: The task's result. Specifically, the fitted estimator/pipeline.
    :rtype: dict
    """

    if isinstance(self.train_data_set, str):
      data_set_data_frame = pd.read_csv(self.train_data_set, delimiter=self.field_delimiter)
    else:
      data_set_data_frame = self.train_data_set.copy()

      data_set_data_frame.reset_index(inplace=True, drop=True)

    y = data_set_data_frame[self.label_column]

    X = data_set_data_frame.drop(columns=[self.label_column, self.id_column])

    if self.feature_columns != 'all':
      X = X[self.feature_columns]

    np.random.seed(self.random_seed)

    estimator = self._build_estimator()

    estimator.fit(X, y)

    joblib.dump(estimator, os.path.join(output_directory, 'trained_model.pkl'))

    return { 'estimator': estimator }

  def _build_estimator (self):
    estimator = copy.deepcopy(self.estimator)

    if self.estimator_params: estimator.set_params(**self.estimator_params)

    return estimator


import os, copy

import pandas as pd

import numpy as np

from sklearn.model_selection import GridSearchCV, RandomizedSearchCV

from . import BaseCrossValidationTask

class HyperParametersSearchCrossValidationTask(BaseCrossValidationTask):
  """
  The :class:`.HyperParametersSearchCrossValidationTask` class can be used to search the best hyperparameters of a scikit-learn estimator/pipeline on some data.

  **Cross-Validation**

  It can support both stratified k-fold cross-validation as well as cross-validation with user-defined folds.

  By default, stratified k-fold cross-validation is used with the default parameters of :meth:`.stratified_folds` method.

  **Search**

  It can support both grid search as well as random search.

  By default, grid search is used.
  """
  def __init__ (self, estimator, search_params, train_data_set, estimator_params=None, field_delimiter=',', scorers=['roc_auc', 'average_precision', 'f1', 'precision', 'recall', 'accuracy'], feature_columns='all', id_column='id', label_column='label', objective_score='f1', random_seed=42, verbose=3, n_jobs=1, return_train_score=True):
    """
    This is the constructor method and can be used to create a new object instance of :class:`.HyperParametersSearchCrossValidationTask` class.

    :param estimator: It can be either an estimator (e.g., LogisticRegression) or a pipeline ending with an estimator.
    :type estimator: scikit-learn {estimator, pipeline}

    :param search_params: Dictionary with hyperparameters names as keys and lists of hyperparameter settings to try as values, or a list of such dictionaries, in which case the grids spanned by each dictionary in the list are explored. This enables searching over any sequence of hyperparameter settings.
    :type search_params: {dict, list of dictionaries}

    :param train_data_set: The input train data set. It can be either a URL, a disk file path or a pandas DataFrame.
    :type train_data_set: {str, pandas DataFrame}

    :param estimator_params: The parameters to override in the provided estimator/pipeline. It defaults to None.
    :type estimator_params: dict, optional

    :param field_delimiter: The separation delimiter (comma for CSV, tab for TSV, etc.) used in the input train data set file. It defaults to ','.
    :type field_delimiter: str, optional

    :param scorers: Multiple metrics to evaluate the predictions on the hold-out data. Either give a list of (unique) strings or a dict with names as keys and callables as values. The callables should be scorers built using scikit-learn ``make_scorer``. Note that when using custom scorers, each scorer should return a single value. It defaults to ['roc_auc', 'average_precision', 'f1', 'precision', 'recall', 'accuracy'].
    :type scorers: {list, dict}, optional

    :param feature_columns: Either 'all' to use from the input train data set file all the columns or a list of column names to select specific columns. It defaults to 'all'.
    :type feature_columns: {str, list}, optional

    :param id_column: The name of the column in the input train data set file containing the sample IDs. It defaults to 'id'.
    :type id_column: str, optional

    :param label_column: The name of the column in the input train data set file containing the ground truth labels. It defaults to 'label'.
    :type label_column: str, optional

    :param objective_score: The scorer that would be used to find the best hyperparameters for refitting the best estimator/pipeline at the end. It defaults to 'f1'.
    :type objective_score: str, optional

    :param random_seed: The random seed used in the random number generator. It can be used to reproduce the output. It defaults to 42.
    :type random_seed: int, optional

    :param verbose: Controls the verbosity of output. The higher, the more messages. It defaults to 3.
    :type verbose: int, optional

    :param n_jobs: Number of jobs to run in parallel. -1 means using all processors. It defaults to 1.
    :type n_jobs: int, optional

    :param return_train_score: If False, training scores will not be computed and returned. Computing training scores is used to get insights on how different parameter settings impact the overfitting/underfitting trade-off. It defaults to True.
    :type return_train_score: bool, optional
    """

    super(HyperParametersSearchCrossValidationTask, self).__init__(HyperParametersSearchCrossValidationTask.__name__, locals())

    self.grid_search()

    pd.set_option('display.max_colwidth', None)

  def grid_search(self):
    """
    Optional method.

    Use the grid search method when searching the best hyperparameters.

    :return: The object instance itself.
    :rtype: :class:`.HyperParametersSearchCrossValidationTask`
    """

    options = self._filter_arguments(locals())

    self._update_arguments({ 'search_options' : options})

    self._update_arguments({ 'search_method': 'grid' })

    return self

  def random_search(self, n_iters=200):
    """
    Optional method.

    Use the random search method when searching the best hyperparameters.

    :param n_iters: Number of hyperparameter settings that are sampled. ``n_iters`` trades off runtime vs quality of the solution. It defaults to 200.
    :type n_iters: int, optional

    :return: The object instance itself.
    :rtype: :class:`.HyperParametersSearchCrossValidationTask`
    """

    options = self._filter_arguments(locals())

    self._update_arguments({ 'search_options' : options})

    self._update_arguments({ 'search_method': 'random' })

    return self

  def run(self, output_directory):
    """
    Run the task.

    The search results (``search_results``) are stored also in a *search_results.html* file as a static HTML table under the output directory path.

    :param output_directory: The output directory path under which task-related generated files are stored.
    :type output_directory: str
    
    :return: The task's result. Specifically, **1)** ``best_estimator``: The estimator/pipeline that was chosen by the search, i.e. estimator/pipeline which gave best score on the hold-out data. **2)** ``best_params``: The hyperparameters setting that gave the best results on the hold-out data. **3)** ``best_score``: Mean cross-validated score of the ``best_estimator``. **4)** ``search_results``: Metrics measured for each of the hyperparameters setting in the search. **5)** ``best_index``: The index (of the ``search_results``) which corresponds to the best candidate hyperparameters setting.
    :rtype: dict
    """

    if isinstance(self.train_data_set, str):
      self.train_data_set_data_frame = pd.read_csv(self.train_data_set, delimiter=self.field_delimiter)
    else:
      self.train_data_set_data_frame = self.train_data_set.copy()

      self.train_data_set_data_frame.reset_index(inplace=True, drop=True)

    y = self.train_data_set_data_frame[self.label_column]

    X = self.train_data_set_data_frame.drop(columns=[self.label_column, self.id_column])

    if self.feature_columns != 'all':
      X = X[self.feature_columns]

    np.random.seed(self.random_seed)

    search = self._build_search_method(self._build_cv_splits(X, y))

    search.fit(X, y)

    cv_results = pd.DataFrame(search.cv_results_).reset_index()

    cv_results.columns = cv_results.columns.str.replace('_test_', '_validation_')

    cv_results.to_html(os.path.join(output_directory, 'search_results.html'), index=False)

    return { 'best_estimator': search.best_estimator_, 'best_params': search.best_params_, 'best_index': search.best_index_, 'best_score': search.best_score_, 'search_results': cv_results }

  def _build_search_method (self, cv):
    if self.search_method == 'random':
      return RandomizedSearchCV(self._build_estimator(), self.search_params, cv=cv, scoring=self.scorers, refit=self.objective_score, return_train_score=self.return_train_score, n_jobs=self.n_jobs, verbose=self.verbose, n_iter=self.search_options['n_iters'], random_state=self.random_seed)
    else:
      return GridSearchCV(self._build_estimator(), self.search_params, cv=cv, scoring=self.scorers, refit=self.objective_score, return_train_score=self.return_train_score, n_jobs=self.n_jobs, verbose=self.verbose)

  def _build_estimator (self):
    estimator = copy.deepcopy(self.estimator)

    if self.estimator_params: estimator.set_params(**self.estimator_params)

    return estimator


from sklearn.base import BaseEstimator

import numpy as np

class ColumnSelector(BaseEstimator):
    """
    The :class:`.ColumnSelector` class is an implementation of a column selector for scikit-learn pipelines.

    It can be used for manual feature selection to select specific columns from an input data set.

    It can select columns either by integer indices or by names.
    """
    def __init__(self, cols, drop_axis=False):
        """
        This is the constructor method and can be used to create a new object instance of :class:`.ColumnSelector` class.

        :param cols: A non-empty list specifying the columns to be selected. For example, [1, 4, 5] to select the 2nd, 5th, and 6th columns, and ['A','C','D'] to select the columns A, C and D.
        :type cols: list

        :param drop_axis: Can be used to reshape the output data set from (n_samples, 1) to (n_samples) by dropping the last axis. It defaults to False.
        :type drop_axis: bool, optional
        """

        self.drop_axis = drop_axis

        self.cols = cols

    def fit_transform(self, X, y=None):
        """
        Returns a slice of the input data set.

        :param X: Input vectors of shape (n_samples, n_features), where n_samples is the number of samples and n_features is the number of features.
        :type X: {NumPy array, pandas DataFrame, SciPy sparse matrix}

        :param y: Ignored.
        :type y: None

        :return: Subset of the input data set of shape (n_samples, k_features), where n_samples is the number of samples and k_features <= n_features.
        :rtype: {NumPy array, SciPy sparse matrix}
        """

        return self.transform(X=X, y=y)

    def transform(self, X, y=None):
        """
        Returns a slice of the input data set.

        :param X: Input vectors of shape (n_samples, n_features), where n_samples is the number of samples and n_features is the number of features.
        :type X: {NumPy array, pandas DataFrame, SciPy sparse matrix}

        :param y: Ignored.
        :type y: None

        :return: Subset of the input data set of shape (n_samples, k_features), where n_samples is the number of samples and k_features <= n_features.
        :rtype: {NumPy array, SciPy sparse matrix}
        """

        if hasattr(X, 'loc') or hasattr(X, 'iloc'):
            if type(self.cols) == tuple:
                self.cols = list(self.cols)

            types = {type(i) for i in self.cols}

            if len(types) > 1: raise Exception('Elements in `cols` should be all of the same data type.')

            if isinstance(self.cols[0], int):
                t = X.iloc[:, self.cols].values
            elif isinstance(self.cols[0], str):
                t = X.loc[:, self.cols].values
            else:
                raise Exception('Elements in `cols` should be either `int` or `str`.')
        else:
            t = X[:, self.cols]

        if t.shape[-1] == 1 and self.drop_axis:
            t = t.reshape(-1)

        if len(t.shape) == 1 and not self.drop_axis:
            t = t[:, np.newaxis]

        return t

    def fit(self, X, y=None):
        """
        This is a mock method and does nothing.

        :param X: Ignored.
        :type X: None

        :param y: Ignored.
        :type y: None

        :return: The object instance itself.
        :rtype: :class:`.ColumnSelector`
        """

        return self


import smtplib

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from .base_notifier import BaseNotifier

class EmailNotifier(BaseNotifier):
  """
  The :class:`.EmailNotifier` class can be used to send email notifications.
  """
  def __init__ (self, email_subject, sender_account, sender_password, smtp_server, smtp_port, recipients):
    """
    This is the constructor method and can be used to create a new object instance of :class:`.EmailNotifier` class.

    :param email_subject: The subject of the email.
    :type email_subject: str

    :param sender_account: The email account of the sender. For example, 'someone@gmail.com'.
    :type sender_account: str

    :param sender_password: The password of the sender email account.
    :type sender_password: str 

    :param smtp_server: The secured SMTP server of the sender email account. For example, for Gmail is 'smtp.gmail.com'.
    :type smtp_server: str

    :param smtp_port: The port of the secured SMTP server. For example, for Gmail is 465.
    :type smtp_port: int

    :param recipients: The recipients (email addresses) as CSV.
    :type recipients: str
    """
    self.recipients = [ o.strip() for o in recipients.split(',') ]

    self.email_subject = email_subject

    self.sender_account = sender_account

    self.sender_password = sender_password

    self.smtp_server = smtp_server

    self.smtp_port = smtp_port

  def notify(self, message):
    """
    Send the email notification.

    :param message: The notification's message.
    :type message: str
    """

    server = smtplib.SMTP_SSL(self.smtp_server, self.smtp_port)

    server.ehlo()

    server.login(self.sender_account, self.sender_password)

    for recipient in self.recipients:
      email = MIMEMultipart('alternative')

      email['From'] = self.sender_account
      email['To'] = recipient
      email['Subject'] = self.email_subject

      email.attach(MIMEText(message, 'html'))

      text = email.as_string()

      server.sendmail(self.sender_account, recipient, text)

    server.quit()


import json, os

from numpyencoder import NumpyEncoder

class TaskRunner:
  """
  The :class:`.TaskRunner` class is a simplified version (in functionality) of the :class:`.Experiment` class.

  It leaves out all the "experiment" stuff and is focused mostly in the execution and tracking of :class:`.BaseTask` tasks.
  """

  def __init__ (self, output_directory_path):
    """
    This is the constructor method and can be used to create a new object instance of :class:`.TaskRunner` class.

    :param output_directory_path: The output directory path under which task-related generated files are stored.
    :type output_directory_path: str
    """

    self._output_directory_path = output_directory_path

    os.makedirs(self._output_directory_path)

  def run(self, task):
    """
    Run a :class:`.BaseTask` task.

    When running a task, its recorded parameters (e.g., *train_task.params*) and any other task-related generated files are stored under output directory for tracking reasons.

    The task's recorded parameters are in JSON format.

    Lastly, in case an exception occurs, a text file (e.g., *train_task.errors*) is generated under output directory containing the error message.

    :param task: The task to run.
    :type task: :class:`.BaseTask`

    :return: The task's result.
    :rtype: Depends on the ``task`` parameter.
    """

    task_type = task.get_type()

    try:
      self._save_configuration_file(task.get_configuration(), task_type)

      return task.run(self._output_directory_path)
    except Exception as exception:
      self._save_errors_file(exception, task_type)

      raise

  def _save_errors_file(self, exception, task_type):
    with open(os.path.join(self._output_directory_path, f'{task_type}.errors'), 'w') as f: f.write(repr(exception))

  def _save_configuration_file(self, configuration, task_type):
    self._save_dictionary_as_json_file(configuration, os.path.join(self._output_directory_path, f'{task_type}.params'))

  def _save_dictionary_as_json_file(self, dictionary, file_path):
    with open(file_path, 'w') as f: f.write(json.dumps(dictionary, default=lambda o: repr(o), indent=True, cls=NumpyEncoder))